"""
Functions for generating alpha factors.
"""
import random
from typing import Callable

from torch import Tensor

from .alpha import Operation, AlphaFactor
from .operations import (
    InputFeature, TsRank, TsMin, TsMax,
    TsCorrPearson, TsCorrSpearman
)
from .data import AlphaData
from .performance import compute_ic


def generate_random_alpha(
    data: AlphaData,
    max_depth: int = 3,
    operations: list[type[Operation]] = [TsRank, TsMin, TsMax, TsCorrPearson, TsCorrSpearman],
    window_range: tuple[int, int] = (5, 30)
) -> AlphaFactor:
    """Generate random alpha factor.
    
    Args:
        - `data` (AlphaData): Input data
        - `max_depth` (int): Maximum depth of operation tree
        - `operations` (list[type[Operation]]): List of available operations
        - `window_range` (tuple[int, int]): Range for window parameters
    
    Returns:
        - `AlphaFactor`: Generated alpha factor
    """
    def _generate_operation(depth: int) -> Operation:
        if depth >= max_depth:
            # At max depth, only allow input features
            feature_idx = random.randint(0, data.n_features - 1)
            return InputFeature(feature_idx, data.feature_names[feature_idx])
        
        op_class = random.choice(operations)
        
        if op_class == InputFeature:
            feature_idx = random.randint(0, data.n_features - 1)
            return InputFeature(feature_idx, data.feature_names[feature_idx])
        
        elif op_class == TsRank:
            input_op = _generate_operation(depth + 1)
            window = random.randint(*window_range)
            return TsRank(input_op, window)
        
        elif op_class == TsMin:
            input_op = _generate_operation(depth + 1)
            window = random.randint(*window_range)
            return TsMin(input_op, window)
        
        elif op_class == TsMax:
            input_op = _generate_operation(depth + 1)
            window = random.randint(*window_range)
            return TsMax(input_op, window)
        
        elif op_class in [TsCorrPearson, TsCorrSpearman]:
            input_op1 = _generate_operation(depth + 1)
            input_op2 = _generate_operation(depth + 1)
            window = random.randint(*window_range)
            return op_class(input_op1, input_op2, window)
        
        else:
            raise ValueError(f"Unknown operation class: {op_class}")
    
    root = _generate_operation(0)
    return AlphaFactor(root, data.feature_names)


def optimize_alpha_population(
    data: AlphaData,
    returns: Tensor,
    population_size: int = 100,
    n_generations: int = 50,
    mutation_rate: float = 0.1,
    tournament_size: int = 5,
    fitness_fn: Callable[[Tensor, Tensor], float] = compute_ic
) -> list[tuple[AlphaFactor, float]]:
    """Optimize population of alpha factors using genetic algorithm.
    
    Args:
        - `data` (AlphaData): Input data
        - `returns` (Tensor): Forward returns
        - `population_size` (int): Size of population
        - `n_generations` (int): Number of generations
        - `mutation_rate` (float): Probability of mutation
        - `tournament_size` (int): Size of tournament for selection
        - `fitness_fn` (Callable[[Tensor, Tensor], float]): Function to compute fitness of alpha factor
    
    Returns:
        - `list[tuple[AlphaFactor, float]]`: List of (alpha factor, fitness) pairs, sorted by fitness
    """
    # Initialize population
    population = [
        generate_random_alpha(data)
        for _ in range(population_size)
    ]
    
    for generation in range(n_generations):
        # Evaluate fitness
        fitness_scores = [
            fitness_fn(alpha.compute(data.values), returns)
            for alpha in population
        ]
        
        # Sort by fitness
        population_with_fitness = list(zip(population, fitness_scores))
        population_with_fitness.sort(key=lambda x: x[1], reverse=True)
        
        # Select best individuals
        new_population = [pair[0] for pair in population_with_fitness[:population_size // 2]]
        
        # Fill rest of population with tournament selection and mutation
        while len(new_population) < population_size:
            # Tournament selection
            tournament = random.sample(population_with_fitness, tournament_size)
            winner = max(tournament, key=lambda x: x[1])[0]
            
            # Mutation
            if random.random() < mutation_rate:
                new_alpha = generate_random_alpha(data)
            else:
                new_alpha = winner
            
            new_population.append(new_alpha)
        
        population = new_population
    
    # Final evaluation
    final_fitness = [
        fitness_fn(alpha.compute(data.values), returns)
        for alpha in population
    ]
    
    # Sort and return results
    results = list(zip(population, final_fitness))
    results.sort(key=lambda x: x[1], reverse=True)
    return results


from torch import Tensor
from scipy.stats import johnsonsu
from pyriemann.utils.covariance import covariance_mest

from MLTT.utils import CACHE_SIZE
from MLTT.cache import conditional_lru_cache

class JohnsonSU:
    """
    Johnson SU distribution class.

    Args:
        - `mu` (float): Location parameter
        - `sigma` (float): Scale parameter
        - `alpha` (float): Skewness parameter
        - `beta` (float): Shape parameter
    """

    def __init__(self, mu: float, sigma: float, alpha: float, beta: float):
        self.mu = mu
        self.sigma = sigma
        self.alpha = alpha
        self.beta = beta

    @staticmethod
    def fit(x: Tensor, /) -> 'JohnsonSU':
        """
        Fit the Johnson SU distribution to data.

        Args:
            - `x` (Tensor): The data to fit the distribution to

        Returns:
            - JohnsonSU: The fitted distribution
        """
        mu, sigma, alpha, beta = johnsonsu.fit(x)
        return JohnsonSU(mu, sigma, alpha, beta)

    @property
    def params(self) -> tuple:
        return self.mu, self.sigma, self.alpha, self.beta

    def pdf(self, x: Tensor) -> Tensor:
        """
        Compute the pdf of a Johnson SU distribution.

        Args:
            - `x` (Tensor): The points to evaluate the pdf at

        Returns:
            - Tensor: The pdf evaluated at x
        """
        return johnsonsu.pdf(x, self.mu, self.sigma, self.alpha, self.beta)

    def cdf(self, x: Tensor) -> Tensor:
        """
        Compute the cdf of a Johnson SU distribution.

        Args:
            - `x` (Tensor): The points to evaluate the cdf at

        Returns:
            - Tensor: The cdf evaluated at x
        """
        return johnsonsu.cdf(x, self.mu, self.sigma, self.alpha, self.beta)

    def sample(self, size: int) -> Tensor:
        """
        Generate random samples from the Johnson SU distribution.
        
        Args:
            - `size` (int): Number of samples to generate
            
        Returns:
            - Tensor: Random samples from the distribution
        """
        return johnsonsu.rvs(self.mu, self.sigma, self.alpha, self.beta, size=size)


@conditional_lru_cache(maxsize=CACHE_SIZE)
def robust_covariance(X: Tensor, m_estimator: str = 'tyl', q: float = 0, tol: float = 10e-3, max_iter: int = 100) -> Tensor:
    """
    Computes a robust estimate of the covariance matrix of a multivariate
    time-series using a fixed-point algorithm and a M-estimator.

    Args:
        - `X` (Tensor): A multivariate time-series of shape (n_channels, n_times)
        - `m_estimator` (str): The type of M-estimator to use:
            - "hub" for Huber's adaptive trimmed mean
            - "stu" for Student-t's robust mean
            - "tyl" for Tyler's resistant mean
        - `q` (float): A parameter of Huber's M-estimator, determines the trade-off
            between robustness to outliers and closeness to the sample mean.
            If q=0, equivalent to Tyler's estimator, if q=1, equivalent to sample covariance
        - `tol` (float): The tolerance for stopping the fixed-point iteration
        - `max_iter` (int): The maximum number of iterations to perform

    Returns:
        - Tensor: A robust estimate of the covariance matrix of shape (n_channels, n_channels)
    """
    return covariance_mest(X, m_estimator=m_estimator, q=q, tol=tol, n_iter_max=max_iter)



from abc import ABC, abstractmethod

import torch
from numpy import sqrt, nan
from scipy.stats import skew, kurtosis

from warnings import warn
from MLTT.allocation.backtesting import BTResult
from MLTT.utils import CACHE_SIZE
from MLTT.cache import conditional_lru_cache

# Abstract base class for all metrics
class Metric(ABC):
    def __init__(self, **params):
        """
        Initializes the metric with given parameters.

        Args:
            - `params` (dict): Dictionary of parameters needed for the metric calculation
            - `T` (int): Number of periods in a year (365 for daily, 12 for monthly, etc)
        """
        self.params = params

    @abstractmethod
    def calculate(self, backtest_result: BTResult) -> float:
        """
        Abstract method to calculate the metric.

        Args:
            - `backtest_result` (BTResult): A backtest result object

        Returns:
            float: The calculated metric
        """
        pass

    def update_params(self, **new_params):
        """
        Updates the metric's parameters.

        Args:
            - `new_params` (dict): Dictionary of new parameters to update
        """
        self.params.update(new_params)

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Abstract property for the metric's name.

        Returns:
            str: The name of the metric
        """
        pass

    @classmethod
    @abstractmethod
    def format_value(cls, value: float) -> str:
        """
        Abstract method to format the metric value.

        Args:
            - `value` (float): The metric value to format. Example for mean profit: 0.55

        Returns:
            str: The formatted metric value. Example for mean profit: "55%"
        """
        pass

    def __call__(self, backtest_result: BTResult) -> float:
        return self.calculate(backtest_result)


# Concrete implementations of different metrics
class AnnualMean(Metric):
    @property
    def name(self) -> str:
        return f'Annual Mean (Annualization: {self.params.get("T", 1)})'

    def calculate(self, backtest_result: BTResult) -> float:
        T = self.params.get('T', 1)
        return torch.mean(backtest_result.net_change) * T

    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{value * 100:.2f}%'

class MonthlyMean(Metric):
    @property
    def name(self) -> str:
        return f'Monthly Mean (Annualization: {self.params.get("T", 1)})'

    def calculate(self, backtest_result: BTResult) -> float:
        T = self.params.get('T', 1)
        return torch.mean(backtest_result.net_change) * T / 12

    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{value * 100:.2f}%'


class AnnualStd(Metric):
    @property
    def name(self) -> str:
        return f'Annual Standard Deviation (Annualization: {self.params.get("T", 1)})'

    def calculate(self, backtest_result: BTResult) -> float:
        T = self.params.get('T', 1)
        return torch.std(backtest_result.net_change) * sqrt(T)

    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{value * 100:.2f}%'


class AnnualSharpe(Metric):
    @property
    def name(self) -> str:
        return f'Annual Sharpe Ratio (Annualization: {self.params.get("T", 1)}, Risk Free Rate: {self.params.get("risk_free_rate", 0) * 100}%)'

    def calculate(self, backtest_result: BTResult) -> float:
        T = self.params.get('T', 1)
        rf = self.params.get('risk_free_rate', 0)
        mean_return = AnnualMean(T=T).calculate(backtest_result)
        std_dev = AnnualStd(T=T).calculate(backtest_result)
        return (mean_return - rf) / std_dev

    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{value:.2f}'


class DownsideStd(Metric):
    @property
    def name(self) -> str:
        return 'Downside Standard Deviation'

    def calculate(self, backtest_result: BTResult) -> float:
        return torch.std(backtest_result.net_change[backtest_result.net_change < 0])   
    
    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{value * 100:.2f}%'


class AnnualDownsideStd(Metric):
    @property
    def name(self) -> str:
        return f'Annual Downside Standard Deviation (Annualization: {self.params.get("T", 1)})'

    def calculate(self, backtest_result: BTResult) -> float:
        T = self.params.get('T', 1)
        return DownsideStd().calculate(backtest_result) * sqrt(T)
    
    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{value * 100:.2f}%'


class AnnualSortino(Metric):
    @property
    def name(self) -> str:
        return f'Annual Sortino Ratio (Annualization: {self.params.get("T", 1)}, Risk Free Rate: {self.params.get("risk_free_rate", 0) * 100}%)'

    def calculate(self, backtest_result: BTResult) -> float:
        rf = self.params.get('risk_free_rate', 0)
        mean_return = AnnualMean(**self.params).calculate(backtest_result)
        downside_std = AnnualDownsideStd(**self.params).calculate(backtest_result)
        return (mean_return - rf) / downside_std

    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{value:.2f}'


class MaxDrawdown(Metric):
    @property
    def name(self) -> str:
        return 'Maximum Drawdown'

    def calculate(self, backtest_result: BTResult) -> float:
        # Use torch.cummax to get the cumulative maximum values
        running_max, _ = torch.cummax(backtest_result.log_equity, dim=0)
        # Calculate drawdowns as difference between running max and current value
        drawdowns = running_max - backtest_result.log_equity
        # Return the maximum drawdown
        return torch.max(drawdowns)
    
    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{value * 100:.2f}%'


class AverageDrawdown(Metric):
    @property
    def name(self) -> str:
        return 'Average Drawdown'

    def calculate(self, backtest_result: BTResult) -> float:
        # Use torch.cummax to get the cumulative maximum values
        running_max, _ = torch.cummax(backtest_result.log_equity, dim=0)
        # Calculate drawdowns as difference between running max and current value
        drawdowns = running_max - backtest_result.log_equity
        # Return the average drawdown
        return torch.mean(drawdowns)

    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{value * 100:.2f}%'


class AnnualCalmar(Metric):
    @property
    def name(self) -> str:
        return f'Annual Calmar Ratio (Annualization: {self.params.get("T", 1)}, Risk Free Rate: {self.params.get("risk_free_rate", 0) * 100}%)'

    def calculate(self, backtest_result: BTResult) -> float:
        rf = self.params.get('risk_free_rate', 0)
        mean_return = AnnualMean(**self.params).calculate(backtest_result)
        max_dd = MaxDrawdown().calculate(backtest_result)
        return (mean_return - rf) / max_dd

    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{value:.2f}'


class ValueAtRisk(Metric):
    @property
    def name(self) -> str:
        return f'Value at Risk (confidence={self.params.get("var_confidence", 0.95) * 100}%)'

    def calculate(self, backtest_result: BTResult) -> float:
        confidence = self.params.get('var_confidence', 0.95)
        return torch.quantile(backtest_result.net_change, 1 - confidence)

    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{-value*100:.2f}%'


class ConditionalValueAtRisk(Metric):
    @property
    def name(self) -> str:
        return f'Conditional Value at Risk (confidence={self.params.get("var_confidence", 0.95) * 100}%)'

    def calculate(self, backtest_result: BTResult) -> float:
        var = ValueAtRisk(**self.params).calculate(backtest_result)
        tail_losses = backtest_result.net_change[backtest_result.net_change <= var]
        return torch.mean(tail_losses)
    
    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{-value*100:.2f}%'


class MedianDrawdown(Metric):
    @property
    def name(self) -> str:
        return 'Median Drawdown'

    def calculate(self, backtest_result: BTResult) -> float:
        return torch.median(torch.cummax(backtest_result.log_equity, dim=0)[0] - backtest_result.log_equity)

    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{value * 100:.2f}%'


class AnnualAlpha(Metric):
    def __init__(self, benchmark: torch.Tensor, **params):
        super().__init__(**params)
        self.benchmark = benchmark

    @property
    def name(self) -> str:
        return f'Annual Alpha (Annualization: {self.params.get("T", 1)})'

    def calculate(self, backtest_result: BTResult) -> float:
        T = self.params.get('T', 1)
        
        avg = AnnualMean(T=T)
        
        # Calculate annual mean returns for data and benchmark
        annual_mean_data = avg.calculate(backtest_result)
        annual_mean_benchmark = avg.calculate(self.benchmark)
        
        # Calculate and return annual alpha as the difference of annual means
        return annual_mean_data - annual_mean_benchmark

    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{value * 100:.2f}%'


class AnnualBeta(Metric):
    def __init__(self, benchmark: torch.Tensor, **params):
        super().__init__(**params)
        self.benchmark = benchmark

    @property
    def name(self) -> str:
        return f'Annual Beta (Annualization: {self.params.get("T", 1)})'

    def calculate(self, backtest_result: BTResult) -> float:
        if len(backtest_result.net_change) != len(self.benchmark):
            warn("Strategy data and benchmark must have the same length")
            return float('nan')
        covariance = torch.cov(backtest_result.net_change, self.benchmark)[0][1]
        benchmark_variance = torch.var(self.benchmark)
        return covariance / benchmark_variance

    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{value:.2f}'


class UlcerIndex(Metric):
    @property
    def name(self) -> str:
        return 'Ulcer Index'

    def calculate(self, backtest_result: BTResult) -> float:
        drawdowns = torch.cummax(backtest_result.log_equity, dim=0)[0] - backtest_result.log_equity
        return torch.sqrt(torch.mean(drawdowns**2))

    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{value:.2f}'


class Skewness(Metric):
    @property
    def name(self) -> str:
        return 'Skewness'

    def calculate(self, backtest_result: BTResult) -> float:
        return skew(backtest_result.net_change.cpu())

    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{value:.4f}'


class Kurtosis(Metric):
    @property
    def name(self) -> str:
        return 'Excess Kurtosis'

    def calculate(self, backtest_result: BTResult) -> float:
        return kurtosis(backtest_result.net_change.cpu())

    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{value:.4f}'


def format_duration(value: float) -> str:
    """
    Convert duration in seconds to a human-readable string.
    """
    if value < 1/1000:  # Less than 1ms
        return f'{value*1000000:.2f} μs'
    elif value < 1:  # Less than 1s
        return f'{value*1000:.2f} ms'
    elif value < 60:  # Less than 1m
        return f'{value:.2f} s'
    elif value < 3600:  # Less than 1h
        return f'{value/60:.2f} m'
    elif value < 86400:  # Less than 1d
        hours = int(value/3600)
        minutes = int((value % 3600)/60)
        return f'{hours}h {minutes}m'
    else:
        days = int(value/86400)
        hours = int((value % 86400)/3600)
        return f'{days}d {hours}h'
    
def years_to_seconds(years: float) -> float:
    return years * 365 * 24 * 60 * 60


def calculate_drawdown_durations(log_equity: torch.Tensor) -> torch.Tensor:
    cum_max = torch.cummax(log_equity, dim=0)[0]
    drawdowns = cum_max - log_equity
    is_drawdown = drawdowns > 0

    # Находим переходы между состояниями просадки
    diff = torch.diff(is_drawdown.int())
    starts = torch.where(diff == 1)[0] + 1
    ends = torch.where(diff == -1)[0] + 1

    # Обрабатываем незавершенную просадку в конце
    if is_drawdown[-1]:
        ends = torch.cat([ends, torch.tensor([len(is_drawdown)], device=ends.device)])

    # Выравниваем длины массивов и конвертируем в годы
    min_len = min(len(starts), len(ends))
    return (ends[:min_len] - starts[:min_len])

class AverageDrawdownDuration(Metric):
    @property
    def name(self) -> str:
        return 'Average Drawdown Duration'

    def calculate(self, backtest_result: BTResult) -> float:
        T = self.params.get('T', 365)
        durations = calculate_drawdown_durations(backtest_result.log_equity) / T
        return years_to_seconds(torch.mean(durations.float())) if len(durations) > 0 else 0.0

    @classmethod
    def format_value(cls, value: float) -> str:
        return format_duration(value)

class MaxDrawdownDuration(Metric):
    @property
    def name(self) -> str:
        return 'Max Drawdown Duration'
    
    def calculate(self, backtest_result: BTResult) -> float:
        T = self.params.get('T', 365)
        durations = calculate_drawdown_durations(backtest_result.log_equity) / T
        return years_to_seconds(torch.max(durations.float())) if len(durations) > 0 else 0.0

    @classmethod
    def format_value(cls, value: float) -> str:
        return format_duration(value)

class AveragePositionDuration(Metric):
    @property
    def name(self) -> str:
        return 'Average Position Duration'

    def calculate(self, backtest_result: BTResult) -> float:
        """
        Calculate the average duration of positions.
        
        !!! IMPORTANT: data array actually not used in this metric. 
        All calculations are based on positions array from initialization.
        """
        weights = backtest_result.weights
        if weights is None:
            return 0.0
            
        # For multi-asset weights, we'll consider a position active if any asset has non-zero weight
        is_long_position = torch.any(weights > 0, axis=1) if len(weights.shape) > 1 else weights > 0
        is_short_position = torch.any(weights < 0, axis=1) if len(weights.shape) > 1 else weights < 0

        # Find the indices where long positions start and end
        long_starts = torch.where(torch.diff(is_long_position.int()) == 1)[0] + 1
        long_ends = torch.where(torch.diff(is_long_position.int()) == -1)[0] + 1

        # If the last period is a long position, add the end index
        if is_long_position[-1]:
            long_ends = torch.cat([long_ends, torch.tensor([len(is_long_position)], device=long_ends.device)])

        # Find the indices where short positions start and end
        short_starts = torch.where(torch.diff(is_short_position.int()) == 1)[0] + 1
        short_ends = torch.where(torch.diff(is_short_position.int()) == -1)[0] + 1

        # If the last period is a short position, add the end index
        if is_short_position[-1]:
            short_ends = torch.cat([short_ends, torch.tensor([len(is_short_position)], device=short_ends.device)])

        # Calculate long position durations
        long_durations = long_ends - long_starts if len(long_starts) > 0 and len(long_ends) > 0 else torch.tensor([])

        # Calculate short position durations
        short_durations = short_ends - short_starts if len(short_starts) > 0 and len(short_ends) > 0 else torch.tensor([])

        # Combine long and short durations
        all_durations = torch.cat((long_durations, short_durations))

        if len(all_durations) > 0:
            T = self.params.get('T', 365)
            return years_to_seconds(torch.mean(all_durations.float()) / T)
        else:
            return 0.0

    @classmethod
    def format_value(cls, value: float) -> str:
        return format_duration(value)

class DailyTurnover(Metric):
    @property
    def name(self) -> str:
        return f'Daily Turnover (Annualization: {self.params.get("T", 1)})'

    def calculate(self, backtest_result: BTResult) -> float:
        weights = backtest_result.weights  # Shape (n_observations, n_assets)
        if weights is None:
            return nan
        T = self.params.get('T', 1)

        if weights.shape[0] < 2:
            return 0.0  # Not enough data to compute turnover

        # Compute absolute weight changes
        turnover = torch.sum(torch.abs(torch.diff(weights, dim=0)), dim=1)

        # Average turnover over time and annualize
        avg_turnover = torch.mean(turnover)

        return avg_turnover * T / 365
    
    @classmethod
    def format_value(cls, value: float) -> str:
        return f'{value * 100:.2f}%'

ALL_METRICS = [
    AnnualMean(T=365),
    MonthlyMean(T=365),
    AnnualStd(T=365),
    AnnualSharpe(T=365, risk_free_rate=0.02),
    AnnualSortino(T=365, risk_free_rate=0.02),
    AnnualCalmar(T=365, risk_free_rate=0.02),
    MaxDrawdown(),
    AverageDrawdown(),
    MedianDrawdown(),
    ValueAtRisk(var_confidence=0.95),
    ConditionalValueAtRisk(var_confidence=0.95),
    UlcerIndex(),
    Skewness(),
    Kurtosis(),
    AverageDrawdownDuration(T=365),
    MaxDrawdownDuration(T=365),
    AveragePositionDuration(T=365),
    DailyTurnover(T=365)
]



# TODO: SWITCH TO TIMESCALEDB
import numpy as np


def extract_time(data: np.ndarray, symbols: list[str], position: int) -> dict[str, int]:
    """Extracts timestamps from specific position in data array.

    Args:
        data (np.ndarray): Array containing market data
        symbols (list[str]): List of trading pair symbols
        position (int): Index position to extract timestamp from (0 for first, -1 for last)

    Returns:
        dict[str, int]: Dictionary mapping symbols to their timestamps
    """
    result = {}
    for i, start in enumerate(data[:, position, 0]):
        result[symbols[i]] = int(start)
    return result


def get_timestamps(klines: np.ndarray, symbols: list[str]) -> tuple[dict[str, int], dict[str, int]]:
    """Gets first and last timestamps for each trading pair.

    Args:
        klines (np.ndarray): Array containing market data
        symbols (list[str]): List of trading pair symbols

    Returns:
        tuple[dict[str, int], dict[str, int]]: Two dictionaries containing:
            - Last timestamps for each symbol
            - First timestamps for each symbol
    """
    if not isinstance(klines, np.ndarray):
        raise ValueError(
            f"klines must be np.ndarray, but got {type(klines)}."
            "You should use `equals=True` when loading data from files")

    if klines.ndim != 3:
        raise ValueError(f"klines must be 3-dimensional array, but got {klines.ndim}-dimensional")

    last_infos = extract_time(klines, symbols, -1)
    first_infos = extract_time(klines, symbols, 0)

    return last_infos, first_infos


def find_incorrect_symbols(last_infos: dict[str, int],
                         first_infos: dict[str, int]) -> list[str]:
    """Identifies symbols with mismatched timestamps.

    Args:
        last_infos (dict[str, int]): Dictionary of last timestamps
        first_infos (dict[str, int]): Dictionary of first timestamps

    Returns:
        list[str]: List of symbols that have inconsistent timestamps
    """
    starts = []
    ends = []
    incorrect_symbols = []

    for symbol_, timestamp_line in last_infos.items():
        ends.append(int(timestamp_line))
        starts.append(int(first_infos[symbol_]))

    max_end_time = max(ends)
    max_start_time = max(starts)

    for symbol_, timestamp_line in last_infos.items():
        start = int(first_infos[symbol_])
        if timestamp_line != max_end_time or start != max_start_time:
            incorrect_symbols.append(symbol_)

    return incorrect_symbols


def filter_symbols_by_correct_timestamp(klines: np.ndarray,
                                      symbols: list[str]) -> list[str]:
    """Filters out symbols with inconsistent timestamps.

    Args:
        klines (np.ndarray): Array containing market data
        symbols (list[str]): List of trading pair symbols

    Returns:
        list[str]: List of symbols with consistent timestamps
    """
    last_infos, first_infos = get_timestamps(klines, symbols)
    incorrect_symbols = find_incorrect_symbols(last_infos, first_infos)
    correct_symbols = set(symbols) - set(incorrect_symbols)

    return list(correct_symbols)


def select_symbols_by_observations(min_data_length: int,
                                 klines: np.ndarray,
                                 symbols: list[str]) -> list[str]:
    """Selects symbols that have enough historical data.

    Args:
        min_data_length (int): Minimum required number of observations
        klines (np.ndarray): Array containing market data
        symbols (list[str]): List of trading pair symbols

    Returns:
        list[str]: List of symbols that meet the minimum data requirement
    """
    correct_symbols = []

    for symbol, kline in zip(symbols, klines):
        if kline.shape[0] >= min_data_length:
            correct_symbols.append(symbol)

    return correct_symbols


def max_information_length(depth_curve: list[int],
                         information_criterion=None) -> int:
    """Determines optimal data length based on information criterion.

    Args:
        depth_curve (list[int]): List of available data lengths
        information_criterion (callable, optional): Custom function to evaluate information content.
            If None, uses default criterion: depth_curve[i]*(n_pairs-i)

    Returns:
        int: Optimal data length that maximizes information content
    """
    n_pairs = len(depth_curve)

    if information_criterion is None:
        information_criterion = lambda i: depth_curve[i]*(n_pairs-i)

    informs = [information_criterion(i) for i in range(n_pairs)]

    return depth_curve[np.argmax(informs)]

def frames_lengths(klines: list[np.ndarray]) -> list[int]:
    """Gets lengths of all data frames.

    Args:
        klines (list[np.ndarray]): List of arrays containing market data

    Returns:
        list[int]: List of lengths for each array
    """
    return [klines[i].shape[0] for i in range(len(klines))]


def filter_symbols_by_maximum_information(klines: list[np.ndarray],
                                        symbols: list[str],
                                        information_criterion=None) -> list[str]:
    """Filters symbols to maximize information content while maintaining data consistency.

    Args:
        klines (list[np.ndarray]): List of arrays containing market data
        symbols (list[str]): List of trading pair symbols
        information_criterion (callable, optional): Custom function to evaluate information content

    Returns:
        list[str]: Filtered list of symbols that optimize information content
    """
    lengths = frames_lengths(klines)

    min_data_length = max_information_length(sorted(lengths), information_criterion)
    filtered_symbols = select_symbols_by_observations(min_data_length, klines, symbols)

    return filtered_symbols


def prune_klines(klines: list[np.ndarray]) -> np.ndarray:
    """Combines multiple kline arrays into single array with consistent length.

    Args:
        klines (list[np.ndarray]): List of arrays containing market data

    Returns:
        np.ndarray: Combined array with shape (n_symbols, min_length, n_features)

    Note:
        This operation requires approximately 2x the memory of input data
    """
    min_len = min(kline.shape[0] for kline in klines)
    dst = np.empty((len(klines), min_len, klines[0].shape[1]), dtype=klines[0].dtype)
    for i, kline in enumerate(klines):
        dst[i, -kline.shape[0]:] = kline[-min_len:]
        klines[i] = None
    return dst


def remove_symbols(klines: np.ndarray, symbols: list[str], correct_symbols: list[str]):
    """Removes all symbols from `klines` that are not in `correct_symbols`.

    Args:
        `klines` (np.ndarray): Data array with k-lines information. Shape: `(n_symbols, n_observations, n_features)`
        `symbols` (list[str]): Symbols or identifiers for each data row. Length: `n_symbols`
        `correct_symbols` (list[str]): List of symbols that should be present in `klines` after removal. Length: `<=n_symbols`
    """
    for i in range(len(symbols)-1, -1, -1):
        if symbols[i] not in correct_symbols:
            del klines[i]
            symbols.pop(i)

def make_klines_array(klines: list[np.ndarray],
                      symbols: list[str],
                      information_criterion=None) -> tuple[np.ndarray, list[str]]:
    """Creates unified array from multiple kline arrays, ensuring data consistency.

    Args:
        klines (list[np.ndarray]): List of arrays containing market data
        symbols (list[str]): List of trading pair symbols
        information_criterion (callable, optional): Custom function to evaluate information content

    Returns:
        tuple[np.ndarray, list[str]]: Tuple containing:
            - Combined array with consistent timestamps and optimal information content
            - List of symbols included in the final array
    """
    corrected_symbols = filter_symbols_by_maximum_information(klines, symbols, information_criterion)
    remove_symbols(klines, symbols, corrected_symbols)
    symbols = corrected_symbols
    klines = prune_klines(klines)
    corrected_symbols = filter_symbols_by_correct_timestamp(klines, symbols)
    indices = [symbols.index(symbol) for symbol in corrected_symbols]
    return klines[indices], corrected_symbols



import asyncio
from defillama import DefiLlama as dfl
import pandas as pd
import requests
from json.decoder import JSONDecodeError


def _get_chains(url: str = "https://raw.githubusercontent.com/DefiLlama/DefiLlama-Adapters/refs/heads/main/projects/helper/chains.json"):
    response = requests.get(url)
    return response.json()


def _format_dfl_frame(data: list[dict], 
                      date_index: bool = True,
                      format_date: bool = True) -> pd.DataFrame:
    """
    Format the data from DefiLlama into a pandas DataFrame.


    Example input:
    [
        {'date': 1524787200, 'tvl': 404078},
        {'date': 1524873600, 'tvl': 400585},
        {'date': 1524960000, 'tvl': 429214},
        {'date': 1525046400, 'tvl': 421651},
        ...
    ]
    """
    if date_index:
        frame = pd.DataFrame(data).set_index("date")
    else:
        frame = pd.DataFrame(data)

    if format_date:
        if date_index:
            frame.index = pd.to_datetime(frame.index, unit="s")
        else:
            frame["date"] = pd.to_datetime(frame["date"], unit="s")
    return frame

def _format_slash_nan(value: str, nan_value = pd.NA) -> str:
    """
    Convert "-" to NaN.
    """
    return nan_value if value == "-" else value


class DefiLlamaAdapter:
    def __init__(self):
        self.dfl = dfl()

    def market_tvl(self) -> pd.DataFrame:
        """
        Get historical total value locked (TVL) data for a entire DeFi market from DefiLlama.

        Returns:
            - `pd.DataFrame`: DataFrame with date index and columns:
                - `tvl`: Total Value Locked
        """
        return _format_dfl_frame(
            self.dfl.get_historical_tvl()
        ).rename(columns={"tvl": "mkt_tvl"})

    def _protocols(self) -> pd.DataFrame:
        """
        Get all protocols from DefiLlama.
        """
        data = self.dfl.get_all_protocols()
        protocols_list = []
        for protocol in data:
            protocols_list.append({
                'symbol': _format_slash_nan(protocol['symbol']),
                'name': _format_slash_nan(protocol['name']),
                'slug': _format_slash_nan(protocol['slug']),
                'category': _format_slash_nan(protocol['category']),
                'chains': _format_slash_nan(protocol['chains'])
            })
        return pd.DataFrame(protocols_list)

    async def _chain_tvl(self, chain: str) -> pd.DataFrame:
        """
        Get historical total value locked (TVL) data for a given chain from DefiLlama.
        """
        try:
            response = self.dfl.get_historical_tvl_chain(chain)
            if isinstance(response, dict) and response.get("statusCode") != 200:  # Check HTTP status code first
                return pd.DataFrame()
            return _format_dfl_frame(response)
        except (JSONDecodeError, requests.exceptions.RequestException) as e:
            print(f"Error fetching data for chain {chain}: {str(e)}")
            return pd.DataFrame()

    async def chain_tvl(
            self, 
            chain: str | list[str] | None = None) -> pd.DataFrame:
        """
        Get historical total value locked (TVL) data for given chain(s).

        Args:
            - `chain` (str | list[str] | None): The chain(s) to get TVL data for. 
                Default is None, which will return all chains.
        
        Returns:
            - `pd.DataFrame`: DataFrame with date index and token 
                columns in format `<chain>.tvl`
        """
        if chain is None:
            chain = _get_chains()

        if isinstance(chain, str):
            chain = [chain]

        for i, c in enumerate(chain):
            chain[i] = c.replace(" ", "-")  # DefiLlama uses "-" for spaces in API
    
        tasks = [self._chain_tvl(c) for c in chain]
        results = await asyncio.gather(*tasks)
        result = [r.rename(columns={"tvl": f"{c}.tvl"}) for c, r in zip(chain, results)]
        return pd.concat(result, axis=1)

    def async_chain_tvl(self, chain: str | list[str] | None = None) -> pd.DataFrame:
        """
        Synchronous wrapper for `chain_tvl`
        """
        return asyncio.run(self.chain_tvl(chain))


import torch
from scipy.stats import ttest_1samp
from dataclasses import dataclass

from MLTT.allocation import CapitalAllocator
from MLTT import backtest_model
from warnings import warn


def _random_prices(shape: tuple[int, int]) -> torch.Tensor:
    prices_matrix = torch.cumsum(torch.randn(*shape), dim=0)
    prices_matrix += torch.randn(1, shape[1])
    return prices_matrix

class ConsistencyCheckResult:
    """Container for leak check results"""
    __slots__ = ['_has_leaks', '_discrepancies', '_tolerance']
    
    def __init__(self, has_leaks: bool, discrepancies: list[tuple[int, float]], tolerance: float):
        self._has_leaks = has_leaks
        self._discrepancies = tuple(discrepancies)
        self._tolerance = tolerance

    @property
    def has_leaks(self) -> bool:
        """Whether any inconsistencies were detected"""
        return self._has_leaks
        
    @property
    def discrepancies(self) -> tuple[tuple[int, float], ...]:
        """Tuple of (timestamp, weight_diff) pairs"""
        return self._discrepancies
        
    @property
    def tolerance(self) -> float:
        """Configured tolerance threshold used in check"""
        return self._tolerance
        
    @property
    def timestamps(self) -> tuple[int, ...]:
        """Timestamps with detected discrepancies"""
        return tuple(t for t, _ in self._discrepancies)
    def __repr__(self) -> str:
        status = "❌ INCONSISTENT" if self.has_leaks else "✅ CONSISTENT"
        details = (
            f"Threshold: {self.tolerance:.1e}\n"
            f"Total discrepancies: {len(self.discrepancies)}\n"
        )
        if self.has_leaks:
            details += f"First 3 timestamps: {self.timestamps[:3]}\n"
        return f"{status} WEIGHTS\n{details}"


class WeightConsistencyChecker:
    """
    Checks for weight inconsistencies at transition points between time windows.
    Detects data leaks by comparing model's weight allocations at overlapping timestamps.
    
    Args:
        - `prediction_info` (torch.Tensor): Input tensor for predictions (n_samples, n_features)
        - `window_size` (int | None): History window size for predictions. None = use all history
        - `step` (int): Step size between validation points
        - `tolerance` (float): Maximum allowed weight difference threshold
    
    Example:
        >>> checker = WeightConsistencyChecker(data, window_size=50)
        >>> has_leaks, discrepancies = checker.run_check(model)
        >>> checker.print_report(discrepancies)
    """
    
    def __init__(
        self,
        prediction_info: torch.Tensor,
        window_size: int | None = None,
        step: int = 1,
        tolerance: float = 1e-10
    ):
        self.data = prediction_info
        self.window_size = window_size
        self.step = step
        self.tolerance = tolerance

    def _validate_params(self, min_obs: int) -> None:
        """Check input parameters validity"""
        if min_obs >= self.data.shape[0]:
            raise ValueError(
                f"Insufficient data ({self.data.shape[0]}) "
                f"for model's minimum observations ({min_obs})"
            )
            
        if self.window_size and self.window_size <= min_obs + 1:
            warn(f"Window size ({self.window_size}) smaller than model's stability minimum ({min_obs + 2})")

    def _prepare_data(self, t: int) -> tuple[torch.Tensor, torch.Tensor]:
        """Get data windows for timestamp t and t-1"""
        if self.window_size is None:
            return self.data[:t-1], self.data[:t]
            
        start_prev = max(0, t-1 - self.window_size)
        start_curr = max(0, t - self.window_size)
        return (
            self.data[start_prev:t-1],
            self.data[start_curr:t]
        )

    def _compare_predictions(
        self, 
        t: int,
        model: CapitalAllocator
    ) -> tuple[int, float] | None:
        """Compare weight allocations at transition point t"""
        data_prev, data_curr = self._prepare_data(t)
        
        weights_prev = model.predict(data_prev)
        weights_curr = model.predict(data_curr)
        
        if weights_curr.shape[0] < 2 or weights_prev.shape[0] < 1:
            return None
            
        max_diff = torch.max(torch.abs(weights_prev[-1] - weights_curr[-2])).item()
        return (t, max_diff) if max_diff > self.tolerance else None

    def check(self, model: CapitalAllocator) -> ConsistencyCheckResult:
        """Execute consistency check and return result object"""
        min_obs = model.min_observations
        self._validate_params(min_obs)
        
        discrepancies = []
        for t in range(min_obs + 1, self.data.shape[0], self.step):
            if diff := self._compare_predictions(t, model):
                discrepancies.append(diff)
                
        return ConsistencyCheckResult(
            has_leaks=len(discrepancies) > 0,
            discrepancies=discrepancies,
            tolerance=self.tolerance
        )

@dataclass(frozen=True)
class LeakTestResult:
    """Represents results of data leak check using random data"""
    pvalue: float
    data_shape: tuple[int, int] | None = None
    used_external_data: bool = False
    signif: float = 0.05
    
    @property
    def has_leaks(self) -> bool:
        """Indicates if potential data leaks were detected"""
        return self.pvalue < self.signif
    
    def __repr__(self) -> str:
        status = "🔴 LEAKS DETECTED" if self.has_leaks else "🟢 NO LEAKS"
        return (
            f"{status}\n"
            f"Significance level: {self.signif:.3f}\n"
            f"P-value: {self.pvalue:.4f}\n"
            f"Data shape: {self.data_shape or 'external_data'}\n"
        )

class RandomLeakTester:
    """
    Performs data leak validation using random/supplied market data
    
    Args:
        - `model` (CapitalAllocator): Strategy model to validate
        - `data_shape` (tuple[int, int] | None): Optional shape for generated data (n_samples, n_assets)
        - `external_data` (torch.Tensor | None): Optional pre-generated data tensor to use instead of random
        - `prediction_info` (torch.Tensor | None): Prediction data template for model input
        - `signif` (float): Statistical test threshold (default: 0.05)
    """
    
    def __init__(
        self,
        model: CapitalAllocator,
        data_shape: tuple[int, int] | None = None,
        external_data: torch.Tensor | None = None,
        prediction_info: torch.Tensor | None = None,
        signif: float = 0.05
    ):
        if not 0 < signif < 1:
            raise ValueError("Significance level must be between 0 and 1")
            
        self.signif = signif
        if not data_shape and not external_data:
            raise ValueError("Must provide either data_shape or external_data")
            
        self.model = model
        self.data_shape = data_shape
        self.external_data = external_data
        self.prediction_info = prediction_info
        self._test_result: LeakTestResult | None = None

    def run_test(self) -> LeakTestResult:
        """Execute full leak test pipeline and return results"""
        prices = self._generate_data()
        returns = self._run_backtest(prices)
        return self._analyze_results(returns, prices.shape)

    def _generate_data(self) -> torch.Tensor:
        """Generate random walk data if no external data provided"""
        if self.external_data is not None:
            return self.external_data
        return _random_prices(self.data_shape)

    def _run_backtest(self, prices: torch.Tensor) -> torch.Tensor:
        """Execute backtest and return strategy returns"""
        result = backtest_model(
            self.model,
            prices,
            prediction_info=self.prediction_info,
            commission=0
        )
        return result.gross_change

    def _analyze_results(self, returns: torch.Tensor, data_shape: tuple[int, ...]) -> LeakTestResult:
        """Calculate statistics and generate final report"""
        returns_np = returns.cpu().numpy() if isinstance(returns, torch.Tensor) else returns
        pvalue = ttest_1samp(returns_np, 0).pvalue
        
        return LeakTestResult(
            pvalue=pvalue,
            data_shape=data_shape,
            used_external_data=self.external_data is not None,
            signif=self.signif
        )

def check_random_data_leaks(
    model: CapitalAllocator,
    data_shape: tuple[int, int] = (1000, 20),
    external_data: torch.Tensor | None = None,
    prediction_info: torch.Tensor | None = None,
    signif: float = 0.05
) -> LeakTestResult:
    """
    Validate strategy for data leaks using random/supplied market data.
    
    Args:
        - `model` (CapitalAllocator): Strategy model to validate
        - `data_shape` (tuple[int, int]): Optional shape for generated data (n_samples, n_assets)
        - `external_data` (torch.Tensor | None): Optional pre-generated data tensor to use instead of random
        - `prediction_info` (torch.Tensor | None): Prediction data template for model input
        - `signif` (float): Statistical significance threshold (default: 0.05)
        
    Returns:
        LeakTestResult: Results of the leak test including p-value and assessment
    """
    return RandomLeakTester(
        model=model,
        data_shape=data_shape,
        external_data=external_data,
        prediction_info=prediction_info,
        signif=signif
    ).run_test()



from MLTT.allocation import (
    BaseAllocator,
    backtest_model,
    backtest,
    BlendingModel,
)
from MLTT.data_loading import (
    load_and_format,
    load_csv_files_from_directory,
    ccxt_ohlcv,
)
from MLTT.data_loading.defillama import DefiLlamaAdapter
from allocation_o2 import CapitalAllocator
from MLTT.cache import cache_mode, CacheMode, conditional_lru_cache
from MLTT.utils import to_weights_matrix, to_weights, change, multivariate_barrier, univariate_barrier


import torch
from torch import Tensor

def compute_ic(alpha: Tensor, returns: Tensor, method: str = "spearman") -> float:
    """Compute Information Coefficient between alpha and returns.
    
    Args:
        - `alpha` (Tensor): Alpha factor values
        - `returns` (Tensor): Forward returns
        - `method` (str): Correlation method ("pearson" or "spearman")
    
    Returns:
        - `float`: Information Coefficient value
    """
    if method == "spearman":
        alpha_ranks = torch.argsort(torch.argsort(alpha)).float()
        return_ranks = torch.argsort(torch.argsort(returns)).float()
        
        # Center the ranks
        alpha_ranks = alpha_ranks - alpha_ranks.mean()
        return_ranks = return_ranks - return_ranks.mean()
        
        # Compute correlation
        num = (alpha_ranks * return_ranks).sum()
        den = torch.sqrt((alpha_ranks * alpha_ranks).sum() * (return_ranks * return_ranks).sum())
        
        return (num / (den + 1e-8)).item()
    
    elif method == "pearson":
        # Center the values
        alpha_centered = alpha - alpha.mean()
        returns_centered = returns - returns.mean()
        
        # Compute correlation
        num = (alpha_centered * returns_centered).sum()
        den = torch.sqrt((alpha_centered * alpha_centered).sum() * (returns_centered * returns_centered).sum())
        
        return (num / (den + 1e-8)).item()
    
    else:
        raise ValueError(f"Unknown correlation method: {method}")



"""
Visualization tools for alpha factors.
"""
from typing import Any
from dataclasses import fields

from graphviz import Digraph

from .base import Operation, AlphaFactor
from .operations import (
    InputFeature, TsRank, TsMin, TsMax,
    TsCorrPearson, TsCorrSpearman
)


def get_operation_style(op: Operation) -> dict[str, Any]:
    """Get visualization style for operation."""
    styles = {
        InputFeature: {
            'shape': 'oval',
            'style': 'filled',
            'fillcolor': '#E8F8F5',
            'fontname': 'Helvetica'
        },
        TsRank: {
            'shape': 'box',
            'style': 'filled',
            'fillcolor': '#D4E6F1',
            'fontname': 'Helvetica'
        },
        TsMin: {
            'shape': 'box',
            'style': 'filled',
            'fillcolor': '#D5F5E3',
            'fontname': 'Helvetica'
        },
        TsMax: {
            'shape': 'box',
            'style': 'filled',
            'fillcolor': '#FADBD8',
            'fontname': 'Helvetica'
        },
        TsCorrPearson: {
            'shape': 'box',
            'style': 'filled',
            'fillcolor': '#FCF3CF',
            'fontname': 'Helvetica'
        },
        TsCorrSpearman: {
            'shape': 'box',
            'style': 'filled',
            'fillcolor': '#F5EEF8',
            'fontname': 'Helvetica'
        }
    }
    return styles.get(type(op), {
        'shape': 'box',
        'fontname': 'Helvetica'
    })


def get_operation_label(op: Operation) -> str:
    """Get label for operation node."""
    if isinstance(op, InputFeature):
        return f"Input\\n{op.feature_name}"
    
    # Get operation name without 'Ts' prefix
    op_name = type(op).__name__[2:]
    
    # Add window size for operations that have it
    if hasattr(op, 'window'):
        return f"{op_name}\\nwindow={op.window}"
    
    return op_name


def get_child_operations(op: Operation) -> list[Operation]:
    """Get all child operations of an operation node."""
    children = []
    for field in fields(op):
        value = getattr(op, field.name)  # Use field.name to get the field name
        if isinstance(value, Operation):
            children.append(value)
        elif isinstance(value, (list, tuple)):
            children.extend(x for x in value if isinstance(x, Operation))
    return children


def visualize_alpha(
    alpha: AlphaFactor,
    filename: str | None = None,
    format: str = 'png',
    view: bool = True
) -> Digraph:
    """Visualize alpha factor as a tree using Graphviz.
    
    Args:
        - `alpha` (AlphaFactor): Alpha factor to visualize
        - `filename` (str | None): Output filename (without extension)
        - `format` (str): Output format ('png', 'pdf', 'svg', etc.)
        - `view` (bool): Whether to open the generated visualization
        
    Returns:
        - `Graphviz Digraph` object
    """
    # Create graph
    dot = Digraph(comment='Alpha Factor Tree')
    dot.attr(rankdir='TB')
    
    # Counter for unique node IDs
    node_counter = 0
    
    def add_operation(op: Operation) -> str:
        """Add operation node to graph and return its ID."""
        nonlocal node_counter
        node_id = f"node{node_counter}"
        node_counter += 1
        
        # Add node with style
        dot.node(node_id, get_operation_label(op), **get_operation_style(op))
        
        # Add edges to child operations
        for child in get_child_operations(op):
            child_id = add_operation(child)
            dot.edge(node_id, child_id)
        
        return node_id
    
    # Build tree
    add_operation(alpha.root)
    
    # Save and show
    if filename:
        dot.render(filename, format=format, view=view, cleanup=True)
    
    return dot


def print_alpha_tree(alpha: AlphaFactor) -> None:
    """Print alpha factor tree in text format.
    
    Args:
        - `alpha` (AlphaFactor): Alpha factor to print
    """
    def print_operation(op: Operation, indent: str = "") -> None:
        """Print operation node."""
        print(f"{indent}└─ {get_operation_label(op)}")
        
        next_indent = indent + "   "
        for child in get_child_operations(op):
            print_operation(child, next_indent)
    
    print("Alpha Factor Tree:")
    print_operation(alpha.root)


def get_mermaid_style(op: Operation) -> str:
    """Get Mermaid style for operation node.
    
    Args:
        - `op` (Operation): Operation to get style for
        
    Returns:
        - `str`: Mermaid style string
    """
    styles = {
        InputFeature: 'style Input fill:#E8F8F5',
        TsRank: 'style Rank fill:#D4E6F1',
        TsMin: 'style Min fill:#D5F5E3',
        TsMax: 'style Max fill:#FADBD8',
        TsCorrPearson: 'style CorrP fill:#FCF3CF',
        TsCorrSpearman: 'style CorrS fill:#F5EEF8'
    }
    return styles.get(type(op), '')


def get_mermaid_label(op: Operation) -> str:
    """Get Mermaid-compatible label for operation node.
    
    Args:
        - `op` (Operation): Operation to get label for
        
    Returns:
        - `str`: Label string formatted for Mermaid
    """
    if isinstance(op, InputFeature):
        return f"Input<br/>{op.feature_name}"
    
    # Get operation name without 'Ts' prefix
    op_name = type(op).__name__[2:]
    
    # Add window size for operations that have it
    if hasattr(op, 'window'):
        return f"{op_name}<br/>window={op.window}"
    
    return op_name


def visualize_alpha_mermaid(alpha: AlphaFactor) -> str:
    """Generate Mermaid diagram code for alpha factor visualization.
    
    Args:
        - `alpha` (AlphaFactor): Alpha factor to visualize
        
    Returns:
        - `str`: Mermaid diagram code as string
    """
    # Start Mermaid graph
    mermaid = ["graph TD"]
    
    # Counter for unique node IDs
    node_counter = 0
    
    def add_operation(op: Operation) -> str:
        """Add operation node to Mermaid code and return its ID."""
        nonlocal node_counter, mermaid
        node_id = f"node{node_counter}"
        node_counter += 1
        
        # Add node with label
        label = get_mermaid_label(op)
        mermaid.append(f"    {node_id}[\"{label}\"]")
        
        # Add style if defined
        style = get_mermaid_style(op)
        if style:
            mermaid.append(f"    {style}")
        
        # Add edges to child operations
        for child in get_child_operations(op):
            child_id = add_operation(child)
            mermaid.append(f"    {node_id} --> {child_id}")
        
        return node_id
    
    # Build tree
    add_operation(alpha.root)
    
    return "\n".join(mermaid)


def save_mermaid_diagram(
    mermaid_code: str,
    filename: str
) -> None:
    """Save Mermaid diagram code to a file.
    
    Args:
        - `mermaid_code` (str): Mermaid diagram code
        - `filename` (str): Output filename (without extension)
    """
    with open(f"{filename}.mmd", "w") as f:
        f.write(mermaid_code) 


import torch
import torch
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from datetime import datetime, timedelta

from warnings import warn

from MLTT.report.metrics import ALL_METRICS, Metric
from MLTT.allocation.backtesting import BTResult
from IPython.display import Markdown


class Report:
    def __init__(self, values: dict[str, float]):
        """
        Initializes the Report object with a dictionary of metric values.

        Args:
            - `values` (dict[str, float]): A dictionary where keys are metric names (str) and values are metric values (float)
        """
        self.values = values

    def as_dict(self) -> dict[str, float]:
        """
        Returns the metric values as a dictionary.

        Returns:
            dict[str, float]: A dictionary where keys are metric names (str) and values are metric values (float)
        """
        return self.values.copy()

    def as_series(self) -> pd.Series:
        """
        Returns the metric values as a pandas Series.

        Returns:
            pd.Series: A pandas Series where index is metric names (str) and values are metric values (float)
        """
        return pd.Series(self.values)
    
    def as_markdown(self) -> str:
        """
        Returns the metric values as a markdown table.
        
        Returns:
            Markdown: IPython.display.Markdown object containing the formatted table
        """
        return generate_markdown_table(self.as_series())

class StrategyReporter:
    def __init__(self, metrics: list[Metric]):
        """
        Initializes the StrategyReporter object with ordered metrics.

        Args:
            - `metrics` (list[Metric]): A list of metric instances
        """
        self.metrics = metrics

    def compose_report(self, backtest_result: BTResult, mode: str = 'visual') -> Report:
        """
        Composes a dictionary of metric values for the given backtest result.

        Args:
            - `backtest_result` (BTResult): A backtest result object
            - `mode` (str): mode of report composition:
                - `visual` - human readable report
                - `numeric` - report with raw values for each metric

        Returns:
            Report: A Report object containing the metric values
        """
        report = {}
        for metric in self.metrics:
            value = metric.calculate(backtest_result)
            if mode == 'visual':
                value = metric.format_value(value)
            report[metric.name] = value
        return Report(report)

    def update_metric_params(self, metric_name: str, **new_params):
        """
        Updates the parameters of a specific metric by name.

        Args:
            - `metric_name` (str): The name of the metric to update
            - `new_params` (dict): Dictionary of new parameters to update
        """
        for metric in self.metrics:
            if metric.name == metric_name:
                metric.update_params(**new_params)
                return
        raise ValueError(f"Metric '{metric_name}' not found in the report.")

    def update_all_metrics_params(self, **new_params):
        """
        Updates parameters for all metrics at once.

        Args:
            - `new_params` (dict): Dictionary of new parameters to update in all metrics
        """
        for metric in self.metrics:
            metric.update_params(**new_params)


def generate_markdown_table(series):
    """
    Generate a markdown-formatted table from a pandas Series.

    Args:
        - `series` (pd.Series): A pandas Series where the index contains metric names and the values
                            are the corresponding metric values

    Returns:
        Markdown: IPython.display.Markdown object containing the formatted table
    """
    # Generate the markdown table header
    markdown_table = "| Metric | Value |\n|--------|-------|\n"

    # Iterate through all items in the series and add them to the table
    for metric, value in series.items():
        if value is not None:
            markdown_table += f"| {metric:<36} | {value:<9} |\n"

    return Markdown(markdown_table)
    
def plot_graphs(backtest_result: BTResult, 
                start_index: int,
                resample_period: str = "ME",
                figsize: tuple[int, int] = (20, 15),
                height_ratios: list[int] = [3, 1, 1],
                timeframe: timedelta | None = None):
    """
    Create a comprehensive visualization of strategy performance metrics.
    
    Args:
        - `backtest_result` (BTResult): The backtest result object containing performance data
        - `start_index` (int): Starting index for the plots
        - `resample_period` (str): Resampling frequency for monthly data ("ME" for month end)
        - `figsize` (tuple[int, int]): Figure size as (width, height)
        - `height_ratios` (list[int]): Height ratios for the subplots
        - `timeframe` (timedelta | None): Time interval between data points
    """
    sum_equity = backtest_result.log_equity
    running_max = torch.cummax(sum_equity, dim=0).values
    drawdown = running_max - sum_equity

    # Create date index ending today with specified timeframe
    end_date = datetime.now()
    if timeframe is None:
        timeframe = timedelta(days=len(backtest_result.net_change))
    
    equity_series = pd.Series(
        backtest_result.net_change.numpy(),
        index=pd.date_range(end=end_date, periods=len(backtest_result.net_change), freq=timeframe)
    )
    month_profits = equity_series.resample(resample_period).sum()

    plt.style.use('dark_background')
    plt.figure(figsize=figsize)
    gs = gridspec.GridSpec(3, 1, height_ratios=height_ratios)

    plot_equity_and_running_max(gs, sum_equity, running_max, start_index)
    plot_drawdown(gs, drawdown, start_index)
    plot_monthly_profits(gs, month_profits, start_index)

    plt.tight_layout()
    plt.show()

def plot_equity_and_running_max(gs, sum_equity, running_max, start_index):
    """
    Plot the logarithmic equity curve and running maximum.
    
    Args:
        - `gs` (matplotlib.gridspec.GridSpec): Grid specification for subplot arrangement
        - `sum_equity` (torch.Tensor): Logarithmic equity curve data
        - `running_max` (torch.Tensor): Running maximum values
        - `start_index` (int): Starting index for the plot
    """
    plt.subplot(gs[0])
    plt.plot(sum_equity[start_index:], label='Log Equity')
    plt.plot(running_max[start_index:], label='Running Max', linestyle='--')
    plt.title('Equity and Running Max Over Time')
    plt.legend()
    plt.tick_params(axis='x', labelbottom=False)  # Hide x-axis labels

def plot_drawdown(gs, drawdown, start_index):
    """
    Plot the drawdown over time.
    
    Args:
        - `gs` (matplotlib.gridspec.GridSpec): Grid specification for subplot arrangement
        - `drawdown` (torch.Tensor): Drawdown values
        - `start_index` (int): Starting index for the plot
    """
    plt.subplot(gs[1])
    plt.plot(drawdown[start_index:], label='Drawdown', color='red')
    plt.title('Drawdown Over Time')
    plt.legend()
    plt.tick_params(axis='x', labelbottom=False)  # Hide x-axis labels

def plot_monthly_profits(gs, month_profits, start_index):
    """
    Plot monthly profits as a bar chart.
    
    Args:
        - `gs` (matplotlib.gridspec.GridSpec): Grid specification for subplot arrangement
        - `month_profits` (pd.Series): Monthly profit data
        - `start_index` (int): Starting index for the plot
    """
    colors = ['#00ff0069' if profit >= 0 else '#ff000044' for profit in month_profits]
    months = month_profits.index.strftime('%b %Y').tolist()
    plt.subplot(gs[2])
    plt.bar(months[start_index:], month_profits[start_index:], color=colors[start_index:])
    plt.title('Monthly Profits')
    plt.xlabel('Month')
    plt.ylabel('Profit')
    plt.xticks(rotation=45)

def write_report_to_file(filename: str, backtest_result: BTResult, resample_period: str = "ME", timeframe: timedelta | None = None):
    """
    Generate a comprehensive strategy report and write it to a file.
    
    Args:
        - `filename` (str): Path to the output file
        - `backtest_result` (BTResult): Backtest result data to analyze
        - `resample_period` (str): Resampling frequency for monthly data ("ME" for month end)
        - `timeframe` (timedelta | None): Time interval between data points
    """
    with open(filename, mode="w") as file:
        write_general_summary(file, backtest_result)
        write_monthly_summaries(file, backtest_result, resample_period, timeframe)

def write_general_summary(file, backtest_result: BTResult):
    """
    Write a general summary of the strategy's performance.
    
    Args:
        - `file` (file-like object): File object to write to
        - `backtest_result` (BTResult): Backtest result data to analyze
    """
    reporter = StrategyReporter(ALL_METRICS)
    report = reporter.compose_report(backtest_result, mode='visual')
    print("General summary:\n", file=file)
    print(report.as_markdown().data, file=file)

def write_monthly_summaries(file, backtest_result: BTResult, resample_period: str, timeframe: timedelta | None = None):
    """
    Write monthly performance summaries.
    
    Args:
        - `file` (file-like object): File object to write to
        - `backtest_result` (BTResult): Backtest result data to analyze
        - `resample_period` (str): Resampling frequency for monthly data ("ME" for month end)
        - `timeframe` (timedelta | None): Time interval between data points
    """
    # Create date index ending today with specified timeframe
    end_date = datetime.now()
    if timeframe is None:
        timeframe = timedelta(days=len(backtest_result.net_change))
    start_date = end_date - timeframe
    
    equity_series = pd.Series(
        backtest_result.net_change.numpy(),
        index=pd.date_range(end=end_date, periods=len(backtest_result.net_change), freq='D')
    )
    monthly_changes = equity_series.resample(resample_period)

    reporter = StrategyReporter(ALL_METRICS)
    for month, changes in monthly_changes:
        monthly_result = BTResult(
            gross_log_change=torch.tensor(changes.values),
            expenses_log=torch.zeros_like(torch.tensor(changes.values)),
            weights=None
        )
        report = reporter.compose_report(monthly_result, mode='visual')
        print("Month:", month.strftime('%B %Y'), file=file)
        print("Portfolio summary this month:\n", file=file)
        print(report.as_markdown().data, file=file)

def monthly_report(backtest_result: BTResult,
                   filename: str | None = None,
                   resample_period: str = "ME",
                   figsize: tuple[int, int] = (20, 15),
                   timeframe: timedelta | None = None):
    """Generate a complete monthly performance report."""
    start_index = 0
    if filename:
        write_report_to_file(filename, backtest_result, resample_period, timeframe)
    plot_graphs(backtest_result, start_index, resample_period, figsize, timeframe=timeframe)



from MLTT.models.kalman import AdaptiveLinearRegression
from MLTT.models.kalman import AdaptiveLinearRegression
from MLTT.models.cointegration import OneLegIndexSignalModel
from MLTT.models.mean_reversion import (CrossSectionalMeanReversionModel, 
                                        EnhancedCrossSectionalMRModel,
                                        BetaAdjustedMeanReversionModel)
from MLTT.models.momentum import SimpleTimeSeriesMomentumModel, TimeSeriesMomentumModel
from MLTT.models.basic import LeakModel, sign_model
from MLTT.models.common import treshold_change


# TODO: SWITCH TO TIMESCALEDB
import os
import glob
import pandas as pd
from tqdm.auto import tqdm
from multiprocessing import Pool


def format_files(source: str, interval: str, out_path: str, verbose: bool = False, overwrite: bool = False) -> None:
    """Formats downloaded Binance data files into a more convenient structure.

    Args:
        - `source` (str): Path to directory containing raw downloaded data
        - `interval` (str): Timeframe of the data ('1m', '1h', etc.)
        - `out_path` (str): Directory where to save formatted files
        - `verbose` (bool, optional): Whether to print progress messages. 
            Defaults to False.
        - `overwrite` (bool, optional): Whether to overwrite existing files. 
            Defaults to False.
    """
    data_folder = f'{source}/klines'
    output_folder = out_path
    os.makedirs(output_folder, exist_ok=True)

    for asset in tqdm(glob.glob(f'{data_folder}/*/{interval}')):
        asset_name = asset.split('/')[-2]
        output_file = os.path.join(output_folder, f'{asset_name}_{interval}.csv')

        if os.path.exists(output_file) and not overwrite:
            if verbose:
                print(f"Skipping: {output_file}")
            continue

        if verbose:
            print(f"Started: {output_file}")

        with open(output_file, 'w') as f_out:
            f_out.write("OpenTime,Open,High,Low,Close,Volume,CloseTime,QuoteAssetVolume,Trades,TakerBuyBaseAssetVolume,TakerBuyQuoteAssetVolume\n")

            for file in os.listdir(asset):
                file_path = os.path.join(asset, file)
                with open(file_path, 'r') as f_in:
                    next(f_in)  # skip header
                    for line in f_in:
                        line = line.rsplit(',', 1)[0] + '\n'
                        f_out.write(line)

        if verbose:
            print(f"Processed: {output_file}")

def process_file(asset: str) -> None:
    """Sorts a single CSV file by timestamp.

    Args:
        asset (str): Path to CSV file to process
    """
    df = pd.read_csv(asset)
    df = df.sort_values('OpenTime')
    df.to_csv(asset, index=False)

def sort_dates(source: str) -> None:
    """Sorts all CSV files in a directory by timestamp using parallel processing.

    Args:
        - `source` (str): Directory containing CSV files to sort
    """
    assets = glob.glob(f'{source}/*.csv')
    with Pool() as pool:
        list(tqdm(pool.imap(process_file, assets), total=len(assets)))



"""
Module for pricing binary outcome events based on price models.

More about PolyMarket-like predictions
"""

import torch
from abc import ABC, abstractmethod
from ..models.price_models import PriceModel


class BinaryPricer(ABC):
    """
    Base class for binary outcome pricers.
    """
    
    def __init__(self, model: PriceModel, num_paths: int = 10000, time_steps: int = 252):
        """
        :param model: Price model instance for simulations
        :param num_paths: Number of paths to simulate for Monte Carlo
        :param time_steps: Number of time steps in simulation
        """
        self.model = model
        self.num_paths = num_paths
        self.time_steps = time_steps
        self._cached_paths = None
        
    def simulate_paths(self):
        """
        Simulate paths using the price model.
        """
        if self._cached_paths is None:
            self._cached_paths = self.model.simulate_paths(self.num_paths, self.time_steps)
        return self._cached_paths
    
    def clear_cache(self):
        """
        Clear cached paths.
        """
        self._cached_paths = None
        
    @abstractmethod
    def estimate_probability(self, strike: float) -> float:
        """
        Estimate probability of the event.
        
        :param strike: Strike price level
        :return: Probability estimate
        """
        pass


class TouchProbabilityPricer(BinaryPricer):
    """
    Estimates probability of price touching strike before expiration.
    """
    
    def estimate_probability(self, strike: float) -> float:
        paths = self.simulate_paths()
        crosses = torch.any(paths >= strike, axis=0)
        return torch.mean(crosses)


class ExpirationProbabilityPricer(BinaryPricer):
    """
    Estimates probability of price being above strike at expiration.
    """
    
    def estimate_probability(self, strike: float) -> float:
        paths = self.simulate_paths()
        final_prices = paths[-1]
        above_strike = final_prices >= strike
        return torch.mean(above_strike)



"""
Alpha factor generation and evaluation package.

This module provides functionality for generating alpha factors, evaluating their performance,
and visualizing the results. It includes operations for creating alpha factors from strings,
computing information coefficients, and visualizing alpha factors.

Note:
    - This module is currently not ready for production use AT ALL.
"""

from MLTT.alphas.base import Operation, AlphaFactor
from MLTT.alphas.data import AlphaData
from MLTT.alphas.performance import compute_ic
from MLTT.alphas.generation import generate_random_alpha, optimize_alpha_population
from MLTT.alphas.operations import InputFeature, TsRank, TsMin, TsMax, TsCorrPearson, TsCorrSpearman
from MLTT.alphas.visualization import visualize_alpha, print_alpha_tree

__all__ = [
    'Operation',
    'AlphaFactor',
    'AlphaData',
    'compute_ic',
    'generate_random_alpha',
    'optimize_alpha_population',
    'InputFeature',
    'TsRank',
    'TsMin',
    'TsMax',
    'TsCorrPearson',
    'TsCorrSpearman',
    'visualize_alpha',
    'print_alpha_tree',
]



from torch import Tensor, zeros_like, cumsum, roll
from MLTT.allocation.backend import allocation_quote_profit
from MLTT.utils import CACHE_SIZE
from MLTT.allocation.allocators import CapitalAllocator

from MLTT.cache import conditional_lru_cache

class BTResult:
    """
    Represents a backtest result with logarithmic performance metrics
    """
    _expenses_log: Tensor
    _gross_log_change: Tensor
    _net_log_change: Tensor
    _log_equity: Tensor | None
    _gross_equity: Tensor | None
    _weights: Tensor | None

    def __init__(self,
                 gross_log_change: Tensor,
                 expenses_log: Tensor | None = None,
                 weights: Tensor | None = None) -> None:
        """
        Args:
            - `gross_log_change` (Tensor): Logarithmic change of equity.
            - `expenses_log` (Tensor): Logarithmic expenses in range `[0, +inf)`.
            - `weights` (Tensor): Weights of the portfolio.
        """
        if expenses_log is None:
            expenses_log = zeros_like(gross_log_change)
        self._expenses_log = expenses_log
        self._gross_log_change = gross_log_change
        self._net_log_change = gross_log_change - self._expenses_log
        self._weights = weights
        self._log_equity = None  # Lazy initialization
        self._gross_equity = None  # Lazy initialization

    @property
    def log_equity(self) -> Tensor:
        if self._log_equity is None:
            self._log_equity = cumsum(self.net_change, dim=0)
        return self._log_equity
        
    @property
    def gross_equity(self) -> Tensor:
        if self._gross_equity is None:
            self._gross_equity = cumsum(self.gross_change, dim=0)
        return self._gross_equity

    @property
    def expenses_log(self) -> Tensor:
        return self._expenses_log

    @property
    def net_change(self) -> Tensor:
        return self._net_log_change

    @property
    def gross_change(self) -> Tensor:
        return self._gross_log_change

    @property
    def weights(self) -> Tensor | None:
        return self._weights

    def __getitem__(self, key: slice | int) -> 'BTResult':
        """
        Implements Python's slice syntax for BTResult objects.
        Allows using obj[start:end] notation.
        """
        if isinstance(key, int):
            start = key
            end = key + 1
        else:
            start = key.start if key.start is not None else 0
            end = key.stop if key.stop is not None else len(self.gross_change)
            
        sliced = self.__class__(
            self.gross_change[start:end],
            self.expenses_log[start:end],
            self.weights[start:end] if self.weights is not None else None
        )
        if self._log_equity is not None:
            sliced._log_equity = self._log_equity[start:end]
        if self._gross_equity is not None:
            sliced._gross_equity = self._gross_equity[start:end]
        return sliced
    

@conditional_lru_cache(maxsize=CACHE_SIZE)
def backtest(
    weights: Tensor,
    prices: Tensor | None = None,
    prices_change: Tensor | None = None,
    commission: float = 0.0,
    check_weights: bool = True,
    save_weights: bool = False
) -> BTResult:
    """
    Function calculates profit from quote weights without any adjustment in time.

    Args:
        - `weights` (Tensor): Weights matrix of shape: `(batch_size, n_tradable)`.
            Sum of absolute values in each row is `<=1` (negative values used for short positions).
            Needs to be already shifted to the next time step to avoid leakage in backtesting.
        - `prices` (Tensor): Log-prices matrix of shape: `(batch_size, n_tradable)`.
        - `prices_change` (Tensor): change of log-prices values in range `(-inf, +inf)` aka log-return.
            Shape: `(batch_size, n_tradable)`. If this argument is provided,
            the `prices` argument is not necessary and has no effect on the function's
            behavior. This is more efficient because calculating prices from price
            changes is computationally more expensive than just multiplying them.
        - `commission` (float): Commission rate in range `[0, 1]`.
        - `check_weights` (bool): Whether to validate input weights.
            If the weights are invalid, function will issue a warning.
        - `save_weights` (bool): Whether to save predicted weights. If False saves memory.
    
    Returns:
        BTResult: backtest summary with `1-d` equity curve and more accurate percentage change.
    """
    gross_change, commission_abs = allocation_quote_profit(weights, prices, prices_change, commission, check_weights)
    return BTResult(
        gross_log_change=gross_change,
        expenses_log=commission_abs,
        weights=weights if save_weights else None
    )


@conditional_lru_cache(maxsize=CACHE_SIZE)
def backtest_model(
        model: CapitalAllocator,
        prices: Tensor,
        prediction_info: Tensor | None = None,
        prices_change: Tensor | None = None,
        check_weights: bool = True,
        commission: float = 0.0,
        save_weights: bool = False
    ) -> BTResult:
    """
    Function to test a given model on a set of features and price data.
    The function takes a model and inputs, and returns a backtest result.

    Args:
        - `model` (CapitalAllocator): Model to be tested. It gives
            a portfolio distribution on each moment.
            The model should have a `min_observations` attribute,
            which is minimum number of observations in a slice.
        - `prices` (Tensor): The prices used to backtest the model.
            Shape: `(batch_size, n_tradable)`.
        - `prediction_info` (Tensor): Input data for the model, usually prices.
            If not provided, `prices` is used.
            Shape: `(batch_size, *n_information)`.
        - `prices_change` (Tensor): Price changes used to backtest the model.
            Shape: `(batch_size, n_tradable)`.
            Only one of `prices` or `prices_change` is required.
            If both are provided, `prices` is ignored.
        - `check_weights` (bool): Whether to check the weights of the model.
            Defaults to True.
        - `commission` (float): Commission to be paid on each trade.
            Defaults to 0.0.
        - `save_weights` (bool): Whether to save predicted weights. If False saves memory.
            Defaults to False.

    Returns:
        BTResult: The result of the backtesting process.
            The result is a curve of equity and percentage change of equity.

    Note:
        - `batch_size` is a number of observations in input data.
        - `n_tradable` is a number of assets.
        - `n_information` is a number of features.
    """
    if prediction_info is None:
        prediction_info = prices
        
    predictions = model(prediction_info)
    weights = roll(predictions, shifts=1, dims=0)
    weights[0] = 0

    return backtest(weights=weights,
                    prices=prices,
                    prices_change=prices_change,
                    check_weights=check_weights,
                    commission=commission,
                    save_weights=save_weights)



from sklearn.model_selection import train_test_split
from typing import TypeVar

T = TypeVar('T')
V = TypeVar('V')

def split_data(X: T, 
               Y: V, 
               train_size: float = 0.5, 
               val_size: float = 0.25, 
               test_size: float = 0.25, 
               random_state: int | None = None, 
               shuffle: bool = False) -> tuple[T, T, T, V, V, V]:
    """
    Splits X and Y dataframes into train, validation, and test sets.

    Parameters:
        X (pd.DataFrame): Features dataframe.
        Y (pd.DataFrame or pd.Series): Target dataframe/series.
        train_size (float): Proportion of the dataset to include in the train set (default 0.7).
        val_size (float): Proportion of the dataset to include in the validation set (default 0.15).
        test_size (float): Proportion of the dataset to include in the test set (default 0.15).
        random_state (int, optional): Random seed for reproducibility.
        shuffle (bool, optional): Whether to shuffle the data before splitting.

    Returns:
        X_train, X_val, X_test, Y_train, Y_val, Y_test: Split dataframes for train, validation, and test sets.
    """
    
    # Ensure that the split sizes add up to 1.0
    assert train_size + val_size + test_size == 1.0, "The sum of train, val, and test sizes must equal 1.0"
    
    # First, split into train and temp (validation + test) sets
    X_train, X_temp, Y_train, Y_temp = train_test_split(X, Y, train_size=train_size, random_state=random_state, shuffle=shuffle)
    
    # Calculate the relative size of validation and test sets within the temp set
    val_ratio = val_size / (val_size + test_size)
    
    # Split temp into validation and test sets
    X_val, X_test, Y_val, Y_test = train_test_split(X_temp, Y_temp, train_size=val_ratio, random_state=random_state, shuffle=shuffle)
    
    return X_train, X_val, X_test, Y_train, Y_val, Y_test



import torch
import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns
import plotly.graph_objects as go
import plotly.subplots as sp
import math
from scipy.stats import gaussian_kde


from copy import deepcopy

from MLTT.allocation import backtest


def walk_forward_index(start_idx: int,
                       end_idx: int,
                       IS_len: int,
                       OOS_len: int) -> tuple[list[slice], list[slice]]:
    """
    Walk-forward function for index-based data.

    Args:
        - `start_idx` (int): Starting index for the walk-forward validation
        - `end_idx` (int): Ending index for the walk-forward validation
        - `IS_len` (int): Length of the in-sample period (training set)
        - `OOS_len` (int): Length of the out-of-sample period (testing set)
        
    Returns:
        tuple[list[slice], list[slice]]: A tuple containing two lists: in-sample slices and out-sample slices
    """
    step = OOS_len

    IS_ranges = []
    OOS_ranges = []
    current_idx = start_idx

    while current_idx + IS_len + OOS_len <= end_idx:
        IS_end = current_idx + IS_len
        OOS_end = IS_end + OOS_len

        IS_ranges.append(slice(current_idx, IS_end))
        OOS_ranges.append(slice(IS_end, OOS_end))

        current_idx += step

    return IS_ranges, OOS_ranges


def _sub_slice(indices: slice, shift: int):
    start_idx = max(0, indices.start + shift)
    return slice(start_idx, indices.stop + shift, indices.step)

def _sub_indices(indices: slice, shift: int):
    return [_sub_slice(idx, shift) for idx in indices]


def default_backtest(commission=0, subtract_one=False):
    def _bt_shift_pred_change(predictions, prices):
        if subtract_one:
            predictions -= 1
        # Replace NumPy roll with PyTorch roll
        weights = torch.roll(predictions, shifts=1, dims=0)

        return backtest(weights, prices, commission=commission)
    return _bt_shift_pred_change


class MetricDistribution:
    """
    A class to represent a collection of metric distributions and provide
    functionality to plot the distributions using either Matplotlib or Plotly.

    Attributes:
        - `_names` (list[str]): The names of the metrics
        - `_data` (list[list[float]]): The list of observations for each metric

    Methods:
        - `plot(n_rows=1, kde=False, subplots=True, lib='matplotlib')`: 
          Plots the distributions of the metrics using either Matplotlib or Plotly
    """

    def __init__(self,
                 metric_names: list[str],
                 observations: list[list[float]]):
        """
        Initializes the MetricDistribution with metric names and corresponding observations.

        Args:
            - `metric_names` (list[str]): A list of metric names
            - `observations` (list[list[float]]): A list of lists, where each inner list contains observations for the respective metric

        Raises:
            ValueError: If the length of `metric_names` and `observations` is not equal
        """
        if len(metric_names) != len(observations):
            raise ValueError(
                "Length of `metric_names` and `observations`"
                f" parameters should be equal. Got: {len(metric_names)} and {len(observations)}"
            )
        self._names = metric_names
        self._data = observations
        self._n_metrics = len(self._names)

    def plot(self,
             n_rows=1,
             kde=False,
             subplots=True,
             lib="matplotlib",
             kde_sigma: float=5.0):
        """
        Plots the metric distributions in a grid format, either using Matplotlib or Plotly,
        with optional Kernel Density Estimation (KDE).

        Args:
            - `n_rows` (int): The number of rows to organize the plots into (default is 1)
            - `kde` (bool): Whether to include Kernel Density Estimation (KDE) in the plots (default is False)
            - `subplots` (bool): Whether to plot each metric in a separate subplot (default is True)
            - `lib` (str): The plotting library to use. Accepts either 'matplotlib' or 'plotly' (default is 'matplotlib')
            - `kde_sigma` (float): Number of standard deviations to include in the KDE (default is 5.0)

        Raises:
            ValueError: If `lib` is not 'matplotlib' or 'plotly'
        """
        self._n_cols = math.ceil(self._n_metrics / n_rows)

        self._validate_lib(lib)
        self._kde_sigma = kde_sigma

        if lib == "matplotlib":
            self._plot_matplotlib(n_rows, kde, subplots)

        elif lib == "plotly":
            self._plot_plotly(n_rows, kde, subplots)

    def _plot_matplotlib(self, n_rows, kde, subplots):
        if subplots:
            fig, axs = plt.subplots(n_rows,
                                    self._n_cols,
                                    figsize=(self._n_cols * 5, n_rows * 5))
            axs = axs.ravel()  # Flatten the axis array if it's 2D
            for i, ax in enumerate(axs[:self._n_metrics]):
                sns.histplot(self._data[i], bins=20, kde=kde, ax=ax, color='b', alpha=0.7)
                ax.set_title(self._names[i])
                ax.grid(True)
            for j in range(i+1, len(axs)):
                fig.delaxes(axs[j])  # Remove unused subplots
        else:
            plt.figure(figsize=(10, 6))
            for i in range(self._n_metrics):
                sns.histplot(self._data[i], bins=20, kde=kde, label=self._names[i], alpha=0.6)
            plt.title("Metric Distributions")
            plt.grid(True)
            plt.legend()
        plt.tight_layout()
        plt.show()

    def _plot_plotly(self, n_rows, kde, subplots):
        if subplots:
            fig = sp.make_subplots(rows=n_rows, cols=self._n_cols, subplot_titles=self._names)
            for i in range(self._n_metrics):
                row = i // self._n_cols + 1
                col = i % self._n_cols + 1
                # Plot histogram
                fig.add_trace(go.Histogram(x=self._data[i], name=self._names[i], opacity=0.7), row=row, col=col)
                # Plot KDE if enabled
                if kde:
                    kde_values = self._calculate_kde(self._data[i])
                    fig.add_trace(go.Scatter(x=kde_values['x'], y=kde_values['y'], mode='lines', name=f"{self._names[i]} KDE"), row=row, col=col)
        else:
            fig = go.Figure()
            for i in range(self._n_metrics):
                fig.add_trace(go.Histogram(x=self._data[i], name=self._names[i], opacity=0.7))
                if kde:
                    kde_values = self._calculate_kde(self._data[i])
                    fig.add_trace(go.Scatter(x=kde_values['x'], y=kde_values['y'], mode='lines', name=f"{self._names[i]} KDE"))
            fig.update_layout(barmode='overlay', title="Metric Distributions")

        fig.update_layout(
            title="Metric Distributions",
            xaxis_title="Value",
            yaxis_title="Frequency",
            barmode='overlay' if not subplots else None,
            template="plotly_dark"
        )
        fig.show()

    def _validate_lib(self, lib):
        if lib not in ["matplotlib", "plotly"]:
            raise ValueError("Invalid plotting library. Please use 'matplotlib' or 'plotly'.")

    def _calculate_kde(self, data, num_points=1000):
        """
        Helper function to calculate the KDE for a given dataset using SciPy.

        Parameters:
        -----------
        data : list[float]
            The data for which KDE will be calculated.
        num_points : int, optional
            Number of points used for estimating the KDE curve (default is 1000).

        Returns:
        --------
        kde_values : dict
            A dictionary containing the x and y values for the KDE plot.
        """
        # Convert to numpy if data is a PyTorch tensor
        if isinstance(data, torch.Tensor):
            data = data.cpu().numpy()
        
        kde = gaussian_kde(data)

        # Calculate mean and std - use torch if data is tensor, otherwise use numpy
        if isinstance(data, torch.Tensor):
            mean = torch.mean(data)
            std = torch.std(data)
            mean = mean.item()
            std = std.item()
        else:
            mean = sum(data) / len(data)
            std = (sum((x - mean) ** 2 for x in data) / len(data)) ** 0.5

        lower_bound = mean - self._kde_sigma * std
        higher_bound = mean + self._kde_sigma * std

        # Use torch linspace if available, otherwise use numpy
        x_values = torch.linspace(lower_bound, higher_bound, num_points).numpy()
        y_values = kde(x_values)
        return {'x': x_values, 'y': y_values}


class WalkForward:  # TODO: complete refactoring needed. Check log-prices compatibility.
    def __init__(self,
                 model,
                 in_sample_len,
                 out_sample_len,
                 save_models=False,
                 drop='first',
                 train_params=None,
                 model_backtest=None):
        """
        Initialize WalkForward class.

        :param model: Model to be used.
        :param in_sample_len: Length of in-sample period for each training step.
        :param out_sample_len: Length of out-sample period for each validation step.
        :param save_models: Flag to save models after each step.
        :param drop: Strategy to drop excess elements ('first' or 'last').
        :param train_params: Additional parameters for .fit() method.
        :param model_backtest: Adapting function for backtesting the model predictions.
            Should use `predictions` (not shifted) and `prices` arguments.
            Returns percentage change of equity curve
        """
        self.model = model
        self.in_sample_len = in_sample_len
        self.out_sample_len = out_sample_len
        self.save_models = save_models
        self.train_params = train_params if train_params is not None else {}
        self.drop = drop
        self._equities = []

        if not model_backtest:
            model_backtest = default_backtest()

        self._backtester = model_backtest

    def _sample_indices(self, X):
        total_len = len(X)

        n_drop = self.in_sample_len % self.out_sample_len

        if self.drop == 'first':
            start_idx = n_drop
            end_idx = total_len
        elif self.drop == 'last':
            start_idx = 0
            end_idx = total_len - n_drop
        else:
            raise ValueError('Drop strategy should be either "first" or "last".')

        self._start_idx = start_idx
        self._end_idx = end_idx
        self._n_drop = n_drop

        return walk_forward_index(start_idx,
                                  end_idx,
                                  self.in_sample_len,
                                  self.out_sample_len)

    def _validate_input(self, Y):
        if isinstance(Y, torch.Tensor):
            if torch.isnan(Y).any():
                raise ValueError('Y should not contain NaN values.')
        else:
            # Convert to torch tensor if not already
            Y_tensor = torch.tensor(Y.values if hasattr(Y, 'values') else Y)
            if torch.isnan(Y_tensor).any():
                raise ValueError('Y should not contain NaN values.')

    def run(self,
            X, Y,
            model=None,
            shift=0):
        """
        Run walk-forward validation with specified in-sample and out-sample lengths.

        :param X: Features in DataFrame format.
        :param Y: Target variable in DataFrame or Series format.
        :param model: Model to be used.
        :param shift: Shift of the walk-forward validation periods.
        :return: PyTorch tensor or pandas DataFrame. First column contains real values,
                 second column contains predictions of the model.
        """
        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X)
        if not isinstance(Y, pd.DataFrame):
            Y = pd.DataFrame(Y)

        in_sample_indices, out_sample_indices = self._sample_indices(X)

        in_sample_indices = _sub_indices(in_sample_indices, shift)
        out_sample_indices = _sub_indices(out_sample_indices, shift)

        self._validate_input(Y)

        if model is not None:
            self.model = model

        result = torch.full((len(Y),), float('nan'))

        for in_sample_idx, out_sample_idx in zip(in_sample_indices, out_sample_indices):
            X_train, Y_train = X.iloc[in_sample_idx], Y.iloc[in_sample_idx]
            X_test, Y_test = X.iloc[out_sample_idx], Y.iloc[out_sample_idx]

            model = deepcopy(self.model)
            model.fit(X_train, Y_train, **self.train_params)
            pred = model.predict(X_test)
            
            # Convert pred to tensor if it's not already
            if not isinstance(pred, torch.Tensor):
                pred = torch.tensor(pred)
            
            # Fill the result tensor
            result[out_sample_idx] = pred

        return result

    def distributions(self,
                      X, Y,
                      metrics: list[callable],
                      prices: torch.Tensor,
                      n: int = 30,
                      save_equity: bool = True) -> MetricDistribution:
        """
        Runs multiple walk-forward validations with different shifts and computes specified metrics.

        :param metrics: List of callable metrics: BTResult -> float.
        :param n: Number of different shifts.
        :param save_equity: Flag to save percentage change of equity after each shift.
            Data should be saved in `self.equities`
        :return: List of lists. Each sublist contains values of a metric for each shift.
        """
        results = [[] for _ in metrics]
        self._equities = []

        # Using only part of InSample as data model depends
        shift = self.in_sample_len // n

        for i in range(n):
            predictions = self.run(X, Y, shift=i * shift)
            pct_change_equity = self._backtester(predictions, prices)
            for j, metric in enumerate(metrics):
                results[j].append(metric(pct_change_equity))
            if save_equity:
                self._equities.append(pct_change_equity)
        names = [metric.__name__ for metric in metrics]
        return MetricDistribution(names, results)

    @property
    def equities(self):
        return deepcopy(self._equities)



"""
Operations that can be used in alpha factor trees.
"""
from dataclasses import dataclass

import torch
from torch import Tensor
import torch.nn.functional as F

from MLTT.alphas.base import Operation


@dataclass
class InputFeature(Operation):
    """Represents an input feature from the data."""
    
    feature_idx: int
    feature_name: str
    
    def forward(self, x: Tensor) -> Tensor:
        return x[:, self.feature_idx]
    
    def to_string(self) -> str:
        return f"${self.feature_name}"
    
    @classmethod
    def from_string(cls, s: str) -> 'InputFeature':
        assert s.startswith("$")
        name = s[1:]
        # Note: feature_idx will be set by AlphaFactor when deserializing
        return cls(0, name)


@dataclass
class TsRank(Operation):
    """Compute cross-sectional ranks over a rolling window."""
    
    input_op: Operation
    window: int
    
    def forward(self, x: Tensor) -> Tensor:
        values = self.input_op.forward(x)
        n = len(values)
        result = torch.zeros_like(values)
        
        # Compute rolling ranks
        for i in range(self.window - 1, n):
            window_data = values[i - self.window + 1:i + 1]
            result[i] = torch.argsort(torch.argsort(window_data)).float()[-1]
        
        # Fill initial values
        result[:self.window - 1] = result[self.window - 1]
        return result
    
    def to_string(self) -> str:
        return f"TSRANK({self.input_op.to_string()},{self.window})"
    
    @classmethod
    def from_string(cls, s: str) -> 'TsRank':
        assert s.startswith("TSRANK(") and s.endswith(")")
        inner = s[7:-1]
        op_str, window_str = inner.rsplit(",", 1)
        return cls(
            Operation.from_string(op_str),
            int(window_str)
        )


@dataclass
class TsMin(Operation):
    """Compute rolling minimum."""
    
    input_op: Operation
    window: int
    
    def forward(self, x: Tensor) -> Tensor:
        values = self.input_op.forward(x)
        
        # Pad the input to maintain the same length
        pad_size = self.window - 1
        padded_values = F.pad(values.unsqueeze(0).unsqueeze(0), 
                            (pad_size, 0), mode='replicate')
        
        # Apply min pooling
        return F.max_pool1d(-padded_values, self.window, stride=1).squeeze() * -1
    
    def to_string(self) -> str:
        return f"TSMIN({self.input_op.to_string()},{self.window})"
    
    @classmethod
    def from_string(cls, s: str) -> 'TsMin':
        assert s.startswith("TSMIN(") and s.endswith(")")
        inner = s[6:-1]
        op_str, window_str = inner.rsplit(",", 1)
        return cls(
            Operation.from_string(op_str),
            int(window_str)
        )


@dataclass
class TsMax(Operation):
    """Compute rolling maximum."""
    
    input_op: Operation
    window: int
    
    def forward(self, x: Tensor) -> Tensor:
        values = self.input_op.forward(x)
        
        # Pad the input to maintain the same length
        pad_size = self.window - 1
        padded_values = F.pad(values.unsqueeze(0).unsqueeze(0), 
                            (pad_size, 0), mode='replicate')
        
        # Apply max pooling
        return F.max_pool1d(padded_values, self.window, stride=1).squeeze()
    
    def to_string(self) -> str:
        return f"TSMAX({self.input_op.to_string()},{self.window})"
    
    @classmethod
    def from_string(cls, s: str) -> 'TsMax':
        assert s.startswith("TSMAX(") and s.endswith(")")
        inner = s[6:-1]
        op_str, window_str = inner.rsplit(",", 1)
        return cls(
            Operation.from_string(op_str),
            int(window_str)
        )


@dataclass
class TsCorrPearson(Operation):
    """Compute Pearson correlation over a rolling window."""
    
    input_op1: Operation
    input_op2: Operation
    window: int
    
    def forward(self, x: Tensor) -> Tensor:
        values1 = self.input_op1.forward(x)
        values2 = self.input_op2.forward(x)
        
        # Get device of input tensor
        device = x.device
        
        # Ensure values are on the same device
        values1 = values1.to(device)
        values2 = values2.to(device)
        
        # Initialize result tensor
        n = len(values1)
        result = torch.zeros(n, device=device)
        
        # Pad the beginning with the first correlation value
        for i in range(self.window - 1):
            # Use available data for initial windows
            v1 = values1[:i+1]
            v2 = values2[:i+1]
            
            # Center the values
            v1 = v1 - v1.mean()
            v2 = v2 - v2.mean()
            
            # Compute correlation
            num = (v1 * v2).sum()
            den = torch.sqrt((v1 * v1).sum() * (v2 * v2).sum())
            
            result[i] = num / (den + 1e-8)
        
        # Compute rolling correlation for the rest
        for i in range(self.window - 1, n):
            v1 = values1[i - self.window + 1:i + 1]
            v2 = values2[i - self.window + 1:i + 1]
            
            # Center the values
            v1 = v1 - v1.mean()
            v2 = v2 - v2.mean()
            
            # Compute correlation
            num = (v1 * v2).sum()
            den = torch.sqrt((v1 * v1).sum() * (v2 * v2).sum())
            
            result[i] = num / (den + 1e-8)
        
        return result
    
    def to_string(self) -> str:
        return f"TSCORRP({self.input_op1.to_string()},{self.input_op2.to_string()},{self.window})"
    
    @classmethod
    def from_string(cls, s: str) -> 'TsCorrPearson':
        assert s.startswith("TSCORRP(") and s.endswith(")")
        inner = s[8:-1]
        op1_str, rest = inner.split(",", 1)
        op2_str, window_str = rest.rsplit(",", 1)
        return cls(
            Operation.from_string(op1_str),
            Operation.from_string(op2_str),
            int(window_str)
        )


@dataclass
class TsCorrSpearman(Operation):
    """Compute Spearman correlation over a rolling window."""
    
    input_op1: Operation
    input_op2: Operation
    window: int
    
    def forward(self, x: Tensor) -> Tensor:
        values1 = self.input_op1.forward(x)
        values2 = self.input_op2.forward(x)
        
        # Get device of input tensor
        device = x.device
        
        # Ensure values are on the same device
        values1 = values1.to(device)
        values2 = values2.to(device)
        
        # Initialize result tensor
        n = len(values1)
        result = torch.zeros(n, device=device)
        
        # Pad the beginning with the first correlation value
        for i in range(self.window - 1):
            # Use available data for initial windows
            v1 = values1[:i+1]
            v2 = values2[:i+1]
            
            # Convert to ranks
            r1 = torch.argsort(torch.argsort(v1)).float()
            r2 = torch.argsort(torch.argsort(v2)).float()
            
            # Center the ranks
            r1 = r1 - r1.mean()
            r2 = r2 - r2.mean()
            
            # Compute correlation
            num = (r1 * r2).sum()
            den = torch.sqrt((r1 * r1).sum() * (r2 * r2).sum())
            
            result[i] = num / (den + 1e-8)
        
        # Compute rolling correlation for the rest
        for i in range(self.window - 1, n):
            v1 = values1[i - self.window + 1:i + 1]
            v2 = values2[i - self.window + 1:i + 1]
            
            # Convert to ranks
            r1 = torch.argsort(torch.argsort(v1)).float()
            r2 = torch.argsort(torch.argsort(v2)).float()
            
            # Center the ranks
            r1 = r1 - r1.mean()
            r2 = r2 - r2.mean()
            
            # Compute correlation
            num = (r1 * r2).sum()
            den = torch.sqrt((r1 * r1).sum() * (r2 * r2).sum())
            
            result[i] = num / (den + 1e-8)
        
        return result
    
    def to_string(self) -> str:
        return f"TSCORRS({self.input_op1.to_string()},{self.input_op2.to_string()},{self.window})"
    
    @classmethod
    def from_string(cls, s: str) -> 'TsCorrSpearman':
        assert s.startswith("TSCORRS(") and s.endswith(")")
        inner = s[8:-1]
        op1_str, rest = inner.split(",", 1)
        op2_str, window_str = rest.rsplit(",", 1)
        return cls(
            Operation.from_string(op1_str),
            Operation.from_string(op2_str),
            int(window_str)
        )